# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class DbEngine:
    MSSQL = "mssql"
    POSTGRESQL = "postgresql"


class Table:
    USER = "LMSUser"
    SECTION = "LMSSection"
    ASSIGNMENT = "Assignment"
    ASSIGNMENT_SUBMISSION_TYPES = "AssignmentSubmissionType"
    ASSIGNMENT_SUBMISSION = "AssignmentSubmission"
    SECTION_ASSOCIATION = "LMSUserLMSSectionAssociation"
    SECTION_ACTIVITY = "LMSSectionActivity"
    SYSTEM_ACTIVITY = "LMSSystemActivity"
    ATTENDANCE = "LMSUserAttendanceEvent"



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import re
from typing import List, Tuple

import pandas as pd


def _splitter(text: str) -> List[str]:
    if len(text) == 0:
        return []

    if "'" not in text:
        return [text]

    # This regex looks for things like "a" and "b" in `['a', 'b']`.
    return re.findall("'([^']+)'", text)


def split(assignments_df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:

    if assignments_df.empty:
        return (pd.DataFrame(), pd.DataFrame())

    SUB_TYPE = "SubmissionType"

    # These are columns that need to be in the final output, in addition to
    # SubmissionType. Note that "LastModifiedDate" does not make sense in this
    # case - if the source is modified then that value will become a "soft
    # deleted" row instead of being marked as modified.
    SOURCE_COLS = ["SourceSystem", "SourceSystemIdentifier", "CreateDate"]

    # Create a new DataFrame that will hold the Submission Type collection for
    # an Assignment.
    submission_type_df = assignments_df[[SUB_TYPE]].copy()

    # Remove Submission Type from the original DataFrame.
    assignments_df_2 = assignments_df.drop(SUB_TYPE, axis=1)

    # If an assignment has no SubmissionType then the resulting record will have
    # NA as a value. We don't care about those.
    submission_type_df.dropna(inplace=True)

    # No point in further processing if there are no rows.
    if submission_type_df.empty:
        return (assignments_df_2, submission_type_df)

    # Split the "SubmissionType" column, generating a variable number of new
    # columns - based on the number of items in the row that has the most items
    # in it.
    submission_type_df = submission_type_df.apply(
        lambda s: pd.Series(_splitter(s[0])), axis=1
    )

    # Because the operation above preserves the index from the original
    # DataFrame, we can merge the extra source columns from the original
    # DataFrame into the new one.
    submission_type_df[SOURCE_COLS] = assignments_df_2[SOURCE_COLS]

    # We have a variable number of new columns. Create a list containing just
    # the new columns.
    value_vars = list(set(submission_type_df.columns) - set(SOURCE_COLS))

    # The `melt` function takes the `value_var` columns and moves them to rows,
    # with each over the `value_var` column values being loaded into a single
    # new column labeled by `value_name`.
    submission_type_df = submission_type_df.melt(
        id_vars=SOURCE_COLS, value_vars=value_vars, value_name=SUB_TYPE
    )

    # The `melt` function also included the original column name for each
    # `value_var` in a new column called `variable`. Throw this one away.
    submission_type_df.drop(["variable"], inplace=True, axis=1)

    # Let's say that we had data like this in the original DataFrame:
    """
    SourceSystem, SourceSystemIdentifier, ..., SubmissionType,CreateDate,LastModifiedDate
    Canvas, 103, ..., "['online_text_entry', 'online_upload'],2021-03-11,2021-03-12"
    Canvas, 104, ..., "['online_upload'],2021-03-11,2021-03-12"
    """
    # Then the output right now is a DataFrame like this:
    """
    Canvas | 103 | online_text_entry | 2021-03-11
    Canvas | 103 | online_upload     | 2021-03-11
    Canvas | 104 | online_upload     | 2021-03-11
    Canvas | 104 | NA                | 2021-03-11
    """
    # Note the return of "NA". We need to remove these again.
    submission_type_df.dropna(inplace=True)

    return (assignments_df_2, submission_type_df)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from os import path
from typing import List

from sqlalchemy.engine.base import Engine as sa_Engine
from sqlalchemy.exc import ProgrammingError
from sqlparse import split

from edfi_lms_ds_loader.sql_adapter import get_int, execute_statements


logger = logging.getLogger(__name__)

MIGRATION_SCRIPTS = [
    # CAUTION: these scripts will run in order from "top to bottom", so it is
    # critical to maintain the script order at all times.
    "initialize_lms_database",
    "create_processed_files_table",
    "create_user_tables",
    "create_section_tables",
    "create_assignment_tables",
    "create_section_association_tables",
    "create_assignment_submission_tables",
    "create_section_activity_tables",
    "create_system_activity_tables",
    "create_attendance_tables",
    "remove_startdate_enddate_from_sectionassociation"
]


def _get_script_path(engine: sa_Engine, script_name: str) -> str:
    script_dir = path.join(path.dirname(__file__), "scripts", engine.name)
    return path.join(script_dir, script_name)


def _read_statements_from_file(full_path: str) -> List[str]:
    raw_sql: str
    with open(full_path) as f:
        raw_sql = f.read()

    statements: List[str] = split(raw_sql)
    return statements


def _script_has_been_run(engine: sa_Engine, migration: str) -> bool:
    try:
        statement = f"SELECT 1 FROM lms.migrationjournal WHERE script = '{migration}';"
        response = get_int(engine, statement)

        return bool(response == 1)
    except ProgrammingError as error:
        if (
            # PostgreSLQ error
            "psycopg2.errors.UndefinedTable" in error.args[0]
            or
            # SQL Server error
            "Invalid object name" in error.args[0]
        ):
            # This means it is a fresh database where the migrationjournal table
            # has not been installed yet.
            return False

        raise


def _record_migration_in_journal(engine: sa_Engine, migration: str) -> None:
    statement = f"INSERT INTO lms.migrationjournal (script) values ('{migration}');"

    execute_statements(engine, [statement])


def _lms_schema_exists(engine: sa_Engine) -> bool:
    statement = """
select case when exists (
    select 1 from INFORMATION_SCHEMA.SCHEMATA where schema_name = 'lms'
) then 1 else 0 end
""".strip()

    return get_int(engine, statement) == 1


def _run_migration_script(engine: sa_Engine, migration: str) -> None:

    logger.debug(f"Running migration {migration}...")

    migration_script = _get_script_path(engine, f"{migration}.sql")

    statements = _read_statements_from_file(migration_script)
    execute_statements(engine, statements)

    _record_migration_in_journal(engine, migration)

    logger.debug(f"Done with migration {migration}.")


def migrate(engine: sa_Engine) -> None:
    """
    Runs database migration scripts for installing LMS table schema into the
    destination database.

    Parameters
    ----------
    engine: sa_Engine
        SQL Alchemy database engine object.
    """
    logger.info("Begin database auto-migration...")

    if not _lms_schema_exists(engine):
        _run_migration_script(engine, "initialize_lms_database")

    for migration in MIGRATION_SCRIPTS:
        # The following block of code does not belong in _run_migration_script
        # because it will throw an exception if the migration journal does not
        # exist, and therefore is not appropriate when initializing the LMS
        # database.
        if _script_has_been_run(engine, migration):
            logger.debug(
                f"Migration {migration} has already run and will not be re-run."
            )
            continue

        _run_migration_script(engine, migration)

    logger.info("Done with database auto-migration.")



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from dataclasses import dataclass
import os
from typing import List, Union

from configargparse import ArgParser  # type: ignore
from sqlalchemy import create_engine as sa_create_engine
from sqlalchemy.engine import Engine as sa_Engine

from edfi_lms_ds_loader.helpers.constants import DbEngine, LOG_LEVELS
from edfi_lms_ds_loader.mssql_lms_operations import MssqlLmsOperations


@dataclass
class MainArguments:
    """
    Container for holding arguments parsed at the command line.

    Parameters
    ----------
    csv_path : str
        Base path for finding CSV files.
    engine : str
        Database engine, either "mssql" or "postgresql"
    """

    csv_path: str
    engine: str
    log_level: str

    @staticmethod
    def _get_mssql_port(port: Union[int, None]) -> int:
        if not port:
            port = 1433

        return port

    @staticmethod
    def _get_postgresql_port(port: Union[int, None]) -> int:
        if not port:
            port = 5432

        return port

    def set_connection_string_using_integrated_security(
        self, server: str, port: Union[int, None], db_name: str
    ) -> None:
        """
        Creates a PyODBC connection string using integrated security.

        Parameters
        ----------
        server : str
            Database server name or IP address.
        port : int or None
            Database port number.
        db_name : str
            Database name.
        """

        if self.engine == DbEngine.MSSQL:
            port = MainArguments._get_mssql_port(port)
            self.connection_string = f"mssql+pyodbc://{server},{port}/{db_name}?driver=ODBC+Driver+17+for+SQL+Server?Trusted_Connection=yes"
        else:
            raise ValueError(
                f"Invalid `engine` parameter value for integrated database security: {self.engine}"
            )

    def set_connection_string(
        self,
        server: str,
        port: Union[int, None],
        db_name: str,
        username: str,
        password: str,
    ) -> None:
        """
        Creates a PyODBC connection string using username and password.

        Parameters
        ----------
        server : str
            Database server name or IP address.
        port : int or None
            Database port number.
        db_name : str
            Database name.
        username : str
            Database user name.
        password : str
            Database password.
        """

        if self.engine == DbEngine.MSSQL:
            port = MainArguments._get_mssql_port(port)
            self.connection_string = f"mssql+pyodbc://{username}:{password}@{server},{port}/{db_name}?driver=ODBC+Driver+17+for+SQL+Server"
        elif self.engine == DbEngine.POSTGRESQL:
            port = MainArguments._get_postgresql_port(port)
            self.connection_string = (
                f"postgresql://{username}:{password}@{server}:{port}/{db_name}"
            )
        else:
            raise ValueError(f"Invalid `engine` parameter value: {self.engine}")

    def get_db_operations_adapter(self) -> MssqlLmsOperations:
        return MssqlLmsOperations(self.get_db_engine())

    def get_db_engine(self) -> sa_Engine:
        if self.engine == DbEngine.MSSQL:
            return sa_create_engine(self.connection_string)
        raise NotImplementedError(
            f"Support for '{self.engine}' has not yet been implemented."
        )


def parse_main_arguments(args_in: List[str]) -> MainArguments:
    """
    Configures the command-line interface.

    Parameters
    ----------
    args_in : list of str
        Full argument list from the command line.

    Returns
    -------
    arguments  : MainArguments
        A populated `MainArguments` object.
    """

    parser = ArgParser()
    parser.add(  # type: ignore
        "-c",
        "--csvpath",
        help="Base path for input files.",
        required=True,
        env_var="CSV_PATH",
    )
    parser.add(  # type: ignore
        "-e",
        "--engine",
        help="Database engine.",
        choices=[DbEngine.MSSQL, DbEngine.POSTGRESQL],
        default=DbEngine.MSSQL,
        env_var="DB_ENGINE",
    )
    parser.add(  # type: ignore
        "-s",
        "--server",
        help="Database server name or IP address",
        required=True,
        env_var="DB_SERVER",
    )
    parser.add(  # type: ignore
        "--port",
        help="Database server port number",
        type=int,
        env_var="DB_PORT",
    )
    parser.add(  # type: ignore
        "-d",
        "--dbname",
        help="Name of the database with the LMS tables.",
        env_var="DB_NAME",
        required=True,
    )

    USE_INTEGRATED = "--useintegratedsecurity"
    USE_INTEGRATED_SHORT = "-i"

    parser.add(  # type: ignore
        USE_INTEGRATED_SHORT,
        USE_INTEGRATED,
        help="Use Integrated Security for the database connection.",
        action="store_true",
    )
    user_name_required = (
        USE_INTEGRATED not in args_in and USE_INTEGRATED_SHORT not in args_in
    )
    # This parameter doesn't work right when used from a .env file,
    # so adding a manual override
    integrated_env_var = os.getenv("USE_INTEGRATED_SECURITY")
    if integrated_env_var and integrated_env_var.lower() in ("true", "yes", "t", "y"):
        user_name_required = False

    parser.add(  # type: ignore
        "-u",
        "--username",
        required=user_name_required,
        env_var="DB_USERNAME",
        help="Database username, when not using integrated security.",
    )
    parser.add(  # type: ignore
        "-p",
        "--password",
        required=user_name_required,
        env_var="DB_PASSWORD",
        help="Database user password, when not using integrated security.",
    )

    parser.add(  # type: ignore
        "-l",
        "--log-level",
        required=False,
        help="The log level for the tool.",
        choices=LOG_LEVELS,
        type=str,
        default="INFO",
        env_var="LOG_LEVEL",
    )

    args_parsed = parser.parse_args(args_in)
    args_parsed.useintegratedsecurity = (
        args_parsed.useintegratedsecurity or not user_name_required
    )

    arguments = MainArguments(
        args_parsed.csvpath, args_parsed.engine, args_parsed.log_level
    )

    if args_parsed.useintegratedsecurity:
        arguments.set_connection_string_using_integrated_security(
            args_parsed.server, args_parsed.port, args_parsed.dbname
        )
    else:
        arguments.set_connection_string(
            args_parsed.server,
            args_parsed.port,
            args_parsed.dbname,
            args_parsed.username,
            args_parsed.password,
        )

    return arguments



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

# Developer note: this module is deliberately not unit tested. As a facade it is
# by definition complex. In this application, it contains the main application
# logic and is tested simply by running the application. Any branching or
# looping logic should go into other modules where they can be tested easily.

import logging
from os.path import abspath
from typing import Callable, List
from functools import lru_cache

from pandas import DataFrame

from edfi_lms_ds_loader.helpers.argparser import MainArguments
from edfi_lms_ds_loader import migrator
from edfi_lms_file_utils import file_reader, file_repository
from edfi_lms_file_utils.constants import Resources
from edfi_lms_ds_loader import df_to_db
from edfi_lms_ds_loader.mssql_lms_operations import MssqlLmsOperations

logger = logging.getLogger(__name__)


# This module deliberately has no exception handling. Due to the foreign key
# relationships between tables, there is little point to continuing after a
# failure. For example: users upload fails. Then section would work, but most of
# the child-objects of section have a dependency on users. If users succeeds,
# but section fails, then there's little value to having the users and nothing
# else. When we get that far down, better to keep the pattern of not catching
# exceptions throughout then to only catch exceptions on "lower value" targets
# like grades or attendance.

# TODO: consider refactoring for transaction management, so that operations can
# be rolled back. SQL Alchemy transactions can be used with DataFrame, so this
# should be feasible. https://tracker.ed-fi.org/browse/LMS-244


@lru_cache()
def _get_sections_df(csv_path: str) -> DataFrame:
    return file_reader.get_all_sections(csv_path)


@lru_cache()
def _get_assignments_df(csv_path: str) -> DataFrame:
    sections_df: DataFrame = _get_sections_df(csv_path)
    if sections_df.empty:
        return DataFrame()
    return file_reader.get_all_assignments(csv_path, sections_df)


def _get_unprocessed_file_paths(
    db_adapter: MssqlLmsOperations,
    resource_name: str,
    file_paths: List[str],
) -> List[str]:
    processed_files = db_adapter.get_processed_files(resource_name)
    all_paths = set(file_paths)
    return sorted(all_paths - processed_files)


def _upload_files_from_paths(
    db_adapter: MssqlLmsOperations,
    file_paths: List[str],
    resource_name: str,
    read_file_callback: Callable[[str], DataFrame],
    upload_function: Callable[[MssqlLmsOperations, DataFrame], None],
) -> None:
    unprocessed_files: List[str] = _get_unprocessed_file_paths(
        db_adapter, resource_name, file_paths
    )

    for path in unprocessed_files:
        data: DataFrame = read_file_callback(path)
        rows = data.shape[0]
        if rows != 0:
            upload_function(db_adapter, data)
        db_adapter.add_processed_file(path, resource_name, rows)


def _load_users(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    file_paths = file_repository.get_users_file_paths(abspath(csv_path))

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.USERS,
        file_reader.read_users_file,
        df_to_db.upload_users,
    )


def _load_sections(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    file_paths = file_repository.get_sections_file_paths(abspath(csv_path))

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.SECTIONS,
        file_reader.read_sections_file,
        df_to_db.upload_sections,
    )


def _load_assignments(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    sections_df = _get_sections_df(csv_path)
    if sections_df.empty:
        logger.info("No sections loaded. Skipping assignments.")
        return
    sections = set(
        sections_df["SourceSystemIdentifier"]
    )  # we only need to access the last file

    formatted_path = abspath(csv_path)
    file_paths: List[str] = []
    for section in sections:
        file_paths = file_paths + file_repository.get_assignments_file_paths(
            formatted_path, section
        )

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.ASSIGNMENTS,
        file_reader.read_assignments_file,
        df_to_db.upload_assignments,
    )


def _load_attendance_events(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    sections_df: DataFrame = _get_sections_df(csv_path)
    if sections_df.empty:
        logger.info("No sections loaded. Skipping section associations.")
        return
    sections = set(sections_df["SourceSystemIdentifier"])

    formatted_path = abspath(csv_path)
    file_paths: List[str] = []
    for section in sections:
        file_paths = file_paths + file_repository.get_attendance_events_paths(
            formatted_path, section
        )

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.ATTENDANCE_EVENTS,
        file_reader.read_attendance_events_file,
        df_to_db.upload_attendance_events,
    )


def _load_section_associations(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    sections_df: DataFrame = _get_sections_df(csv_path)
    if sections_df.empty:
        logger.info("No sections loaded. Skipping section associations.")
        return
    sections = set(sections_df["SourceSystemIdentifier"])

    formatted_path = abspath(csv_path)
    file_paths: List[str] = []
    for section in sections:
        file_paths = file_paths + file_repository.get_section_associations_file_paths(
            formatted_path, section
        )

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.SECTION_ASSOCIATIONS,
        file_reader.read_section_associations_file,
        df_to_db.upload_section_associations,
    )


def _load_assignment_submissions(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    assignments_df: DataFrame = _get_assignments_df(csv_path)
    if assignments_df.empty:
        logger.info("No assignments loaded. Skipping assignment submissions.")
        return

    formatted_path = abspath(csv_path)
    file_paths: List[str] = []
    for section_id, assignment_id in zip(
        assignments_df.LMSSectionSourceSystemIdentifier,
        assignments_df.SourceSystemIdentifier,
    ):
        file_paths = file_paths + file_repository.get_submissions_file_paths(
            formatted_path,
            section_id,
            assignment_id,
        )

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.SUBMISSIONS,
        file_reader.read_submissions_file,
        df_to_db.upload_assignment_submissions,
    )


def _load_section_activities(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    sections_df: DataFrame = _get_sections_df(csv_path)
    if sections_df.empty:
        logger.info("No sections loaded. Skipping section associations.")
        return
    sections = set(sections_df["SourceSystemIdentifier"])

    formatted_path = abspath(csv_path)
    file_paths: List[str] = []
    for section in sections:
        file_paths = file_paths + file_repository.get_section_activities_file_paths(
            formatted_path, section
        )

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.SECTION_ACTIVITIES,
        file_reader.read_section_activities_file,
        df_to_db.upload_section_activities,
    )


def _load_system_activities(csv_path: str, db_adapter: MssqlLmsOperations) -> None:
    file_paths = file_repository.get_system_activities_file_paths(abspath(csv_path))

    _upload_files_from_paths(
        db_adapter,
        file_paths,
        Resources.SYSTEM_ACTIVITIES,
        file_reader.read_system_activities_file,
        df_to_db.upload_system_activities,
    )


def run_loader(arguments: MainArguments) -> None:
    logger.info("Begin loading files into the LMS Data Store (DS)...")

    migrator.migrate(arguments.get_db_engine())

    csv_path = arguments.csv_path

    db_adapter = arguments.get_db_operations_adapter()

    _load_users(csv_path, db_adapter)
    _load_sections(csv_path, db_adapter)
    # Important: run this immediately after loading sections, before loading other section-related resources
    _load_section_associations(csv_path, db_adapter)
    _load_assignments(csv_path, db_adapter)
    _load_assignment_submissions(csv_path, db_adapter)
    _load_attendance_events(csv_path, db_adapter)
    _load_section_activities(csv_path, db_adapter)
    _load_system_activities(csv_path, db_adapter)

    logger.info("Done loading files into the LMS Data Store.")



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import Callable, List

import pandas as pd

from edfi_lms_ds_loader.helpers.constants import Table
from edfi_lms_ds_loader.mssql_lms_operations import MssqlLmsOperations
from edfi_lms_ds_loader.helpers import assignment_splitter

logger = logging.getLogger(__name__)


def _prepare_staging_table(
    db_adapter: MssqlLmsOperations, df: pd.DataFrame, table: str
) -> None:
    logger.info(f"Uploading {table} file ...")

    db_adapter.disable_staging_natural_key_index(table)
    db_adapter.truncate_staging_table(table)
    db_adapter.insert_into_staging(df, table)
    db_adapter.enable_staging_natural_key_index(table)


def _get_source_system(df: pd.DataFrame) -> str:
    return str(df.iloc[0]["SourceSystem"])


def _upload_assignment_submission_types(
    db_adapter: MssqlLmsOperations, submission_types_df: pd.DataFrame
) -> None:
    TABLE = Table.ASSIGNMENT_SUBMISSION_TYPES

    _prepare_staging_table(db_adapter, submission_types_df, TABLE)

    db_adapter.insert_new_submission_types()

    source_system: str = _get_source_system(submission_types_df)
    db_adapter.unsoft_delete_returned_submission_types(source_system)
    db_adapter.soft_delete_removed_submission_types(source_system)

    logger.info(f"Done with {TABLE} file.")


def upload_file(
    db_adapter: MssqlLmsOperations,
    df: pd.DataFrame,
    table: str,
    db_adapter_insert_method: Callable[[MssqlLmsOperations, str, List[str]], None],
    db_adapter_delete_method: Callable[[MssqlLmsOperations, str, str], None],
) -> None:
    """
    Uploads a DataFrame to the designated LMS table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    df: pd.DataFrame
        A DataFrame to upload.
    table: str
        The destination table.
    db_adapter_insert_method: Callable[[MssqlLmsOperations, str, List[str]], None]
        The MssqlLmsOperations insert method to use for the upload
    db_adapter_delete_method: Callable[[MssqlLmsOperations, str, str], None],
        The MssqlLmsOperations delete method to use for the upload
    """
    if df.empty:
        return

    _prepare_staging_table(db_adapter, df, table)

    columns = list(df.columns)

    db_adapter_insert_method(db_adapter, table, columns)
    db_adapter.copy_updates_to_production(table, columns)
    db_adapter_delete_method(db_adapter, table, _get_source_system(df))

    logger.info(f"Done with {table} file.")


def upload_users(db_adapter: MssqlLmsOperations, users_df: pd.DataFrame) -> None:
    """
    Uploads a User DataFrame to the User table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    users_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        users_df,
        Table.USER,
        MssqlLmsOperations.insert_new_records_to_production,
        MssqlLmsOperations.soft_delete_from_production,
    )


def upload_sections(db_adapter: MssqlLmsOperations, sections_df: pd.DataFrame) -> None:
    """
    Uploads a Section DataFrame to the Section table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    sections_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        sections_df,
        Table.SECTION,
        MssqlLmsOperations.insert_new_records_to_production,
        MssqlLmsOperations.soft_delete_from_production,
    )


def upload_assignments(
    db_adapter: MssqlLmsOperations, assignments_df: pd.DataFrame
) -> None:
    """
    Uploads an Assignments DataFrame to the Assignment and AssignmentSubmissionType
    tables. The specialized function is required in order to work with the
    collection table, AssignmentSubmissionType


    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    df: pd.DataFrame
        A DataFrame to upload.
    """
    if assignments_df.empty:
        return

    assignments_df, submissions_type_df = assignment_splitter.split(assignments_df)

    # Truncate AssignmentDescription to max 1024 characters, matching the database
    assignments_df["AssignmentDescription"] = assignments_df["AssignmentDescription"].astype("str").str[:1024]  # type: ignore
    upload_file(
        db_adapter,
        assignments_df,
        Table.ASSIGNMENT,
        MssqlLmsOperations.insert_new_records_to_production_for_section_relation,
        MssqlLmsOperations.soft_delete_from_production_for_section_relation,
    )

    if not submissions_type_df.empty:
        _upload_assignment_submission_types(db_adapter, submissions_type_df)


def upload_section_associations(
    db_adapter: MssqlLmsOperations, section_associations_df: pd.DataFrame
) -> None:
    """
    Uploads a Section Association DataFrame to the User-Section Association
    table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    section_associations_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        section_associations_df,
        Table.SECTION_ASSOCIATION,
        MssqlLmsOperations.insert_new_records_to_production_for_section_and_user_relation,
        MssqlLmsOperations.soft_delete_from_production_for_section_relation,
    )


def upload_assignment_submissions(
    db_adapter: MssqlLmsOperations, submissions_df: pd.DataFrame
) -> None:
    """
    Uploads an Assignment Submission DataFrame to the Assignment Submission
    table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    submissions_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        submissions_df,
        Table.ASSIGNMENT_SUBMISSION,
        MssqlLmsOperations.insert_new_records_to_production_for_assignment_and_user_relation,
        MssqlLmsOperations.soft_delete_from_production_for_assignment_relation,
    )


def upload_section_activities(
    db_adapter: MssqlLmsOperations, section_activities_df: pd.DataFrame
) -> None:
    """
    Uploads a Section Activity DataFrame to the Section Activity
    table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    section_activities_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        section_activities_df,
        Table.SECTION_ACTIVITY,
        MssqlLmsOperations.insert_new_records_to_production_for_section_and_user_relation,
        MssqlLmsOperations.soft_delete_from_production_for_section_relation,
    )


def upload_system_activities(
    db_adapter: MssqlLmsOperations, system_activities_df: pd.DataFrame
) -> None:
    """
    Uploads a System Activity DataFrame to the System Activity
    table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    system_activities_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        system_activities_df,
        Table.SYSTEM_ACTIVITY,
        MssqlLmsOperations.insert_new_records_to_production_for_user_relation,
        MssqlLmsOperations.soft_delete_from_production,
    )


def upload_attendance_events(
    db_adapter: MssqlLmsOperations, attendance_df: pd.DataFrame
) -> None:
    """
    Uploads a System Activity DataFrame to the System Activity
    table.

    Parameters
    ----------
    db_adapter: MssqlLmsOperations
        Database engine-specific adapter/wrapper for database operations.
    attendance_df: pd.DataFrame
        A DataFrame to upload.
    """
    upload_file(
        db_adapter,
        attendance_df,
        Table.ATTENDANCE,
        MssqlLmsOperations.insert_new_records_to_production_for_attendance_events,
        MssqlLmsOperations.soft_delete_from_production_for_section_relation,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

# Developer note: this adapter module is deliberately not unit tested.

import logging
from typing import Any, Callable, List, TypeVar

from sqlalchemy.engine.base import Engine as sa_Engine
from sqlalchemy.orm import sessionmaker, Session as sa_Session
from sqlalchemy.exc import ProgrammingError


logger = logging.getLogger(__name__)

T = TypeVar('T')


def execute_transaction(
    engine: sa_Engine, function: Callable[[sa_Session], T]
) -> T:
    Session = sessionmaker(bind=engine)

    session = Session()
    response: Any
    try:
        response = function(session)

        session.commit()

        return response
    except ProgrammingError as pe:
        session.rollback()

        # Deliberately bubbling this error up to the stack - but need to make
        # sure it is handled in the standard error log messages as well.
        logger.exception(pe)
        raise
    finally:
        session.close()


def execute_statements(engine: sa_Engine, statements: List[str]) -> None:
    def __callback(session: sa_Session) -> None:
        for statement in statements:
            # Ignore MSSQL "GO" statements
            if statement == "GO":
                continue

            # Deliberately throwing away all results. Counting on exception handling
            # if there are any errors, and migration scripts should not be returning
            # any results.
            session.execute(statement)

    execute_transaction(engine, __callback)


def get_int(engine: sa_Engine, statement: str) -> int:
    def __callback(session: sa_Session) -> int:
        return session.execute(statement).scalar()

    result: int = execute_transaction(engine, __callback)

    if result:
        return result
    else:
        return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edfi_lms_ds_loader', 'edfi_lms_ds_loader.helpers']

package_data = \
{'': ['*'], 'edfi_lms_ds_loader': ['scripts/mssql/*', 'scripts/postgresql/*']}

install_requires = \
['configargparse>=1.2.3,<2.0.0',
 'edfi-lms-extractor-lib==1.0.0',
 'edfi-lms-file-utils==1.0.0',
 'errorhandler>=2.0.1,<3.0.0',
 'pandas>=1.1.2,<2.0.0',
 'psycopg2>=2.8.6,<3.0.0',
 'pyodbc>=4.0.30,<5.0.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'sqlalchemy==1.3.23',
 'sqlparse>=0.4.1,<0.5.0']

setup_kwargs = {
    'name': 'edfi-lms-ds-loader',
    'version': '1.0.0',
    'description': 'Ed-Fi LMS Toolkit Data Store Loader',
    'long_description': '# lms-ds-loader\n\nThe Ed-Fi LMS Data Store Loader is a utility for loading CSV files in the\nLearning Management System Unified Data Model (LMS-UDM) into a Learning\nManagement System Data Store (LMS-DS) database.\n\nThe application processes each file in the input file system by date order, as\nindicated in the file name. If a record is in a file one day, and missing on the\nnext day, then the system "soft deletes" that record by setting the current\ntimestamp into the `deletedat` column. Similarly, if a previously soft deleted record\nreappears later, the record is "un-soft deleted" and updated with any new values.\n\nThis functionality requires that a root level directory only contains files for\none LMS provider. Thus if an education organization uses multiple LMS providers,\nthen each LMS Extractor needs to write files to a separate, dedicated directory,\nand the LMS DS Loader must be run once for each extractor\'s output directory.\n\nLimitations as of April 2021:\n\n* This tool only supports SQL Server (tested on MSSQL 2019). PostgreSQL support will be added at a future data.\n\n## Getting Started\n\n1. SQL Server support requires the Microsoft ODBC 17 driver, which is newer than\n   the ones that come with most operating systems.\n   * Windows: `choco install sqlserver-odbcdriver`\n   * [Linux\n     instructions](https://docs.microsoft.com/en-us/sql/connect/odbc/linux-mac/installing-the-microsoft-odbc-driver-for-sql-server?view=sql-server-ver15)\n     (has not been tested yet)\n1. Requires Python 3.9+ and Poetry.\n1. Install required Python packages:\n\n   ```bash\n   poetry install\n   ```\n\n1. The database account used when running the tool needs to be a member of the\n   following roles in the destination database:\n\n   * db_datareader\n   * db_datawriter\n   * db_ddladmin\n\nNote that the tool automatically manages deployment of the LMS-DS tables into\nthe destination database, under the `lms` schema.\n\n## Running the Tool\n\nFor detailed help, execute `poetry run python edfi_lms_ds_loader -h`.\n\nSample call using full integrated security, loading from the sample files\ndirectory:\n\n```bash\npoetry run python edfi_lms_ds_loader --server localhost --dbname lms_toolkit --useintegratedsecurity --csvpath ../../docs/sample-out\n```\n\n## Developer Notes\n\n### Database-Enabled Integration Tests\n\nThis package contains SQL Server-integrated tests, in addition to the standard\nunit tests. The specialized tests are in the `tests_integration_sql` directory\nand they will run automatically if you call `pytest` with no arguments.\n\nBefore running the tests, create an empty SQL Server database named\n"test_integration_lms_toolkit": `sqlcmd -Q "create database\ntest_integration_lms_toolkit"`. This is required to run the SQL Server\nintegration pytests.\n\nTo run only the _unit tests_: `poetry run pytest tests`. To run only the\nintegration tests, `poetry run pytest tests_integration_sql`.\n\n#### Writing New Tests\n\nPlease note and follow the existing pattern for database-enabled integration\ntests, as this pattern was carefully tuned to support use of database\ntransactions for rolling back changes and thereby leaving behind a clean\ndatabase.\n\n### Dev Operations\n\n1. Style check: `poetry run flake8`\n1. Static typing check: `poetry run mypy .`\n1. Run unit tests: `poetry run pytest`\n1. Run unit tests with code coverage: `poetry run coverage run -m pytest`\n1. View code coverage: `poetry run coverage report`\n\n_Also see\n[build.py](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/docs/build.md)_ for\nuse of the build script.\n\n### Adding New Migrations\n\n1. Create SQL Server and PostgreSQL SQL scripts under\n   `edfi_lms_ds_loader/scripts/<engine name>`, using the same file name for\n   both. Start from the artifact created by MetaEd, and then update the script\n   with these modifications:\n   * Remove the `Id` column and the default constraint for that column.\n   * Add a `DeletedAt` column as a nullable `Datetime2`.\n   * Duplicate the table definition and prefix the duplicate with "stg_".\n   * In the staging table, change the `xyzIdentifier` primary key column name to\n     `StagingId`, and leave out the `DeletedAt` column.\n1. Use `;` (semi-colon) terminators at the end of each SQL statement for both\n   languages. Do not use `GO` in the SQL Server files, as the application is not\n   coded to parse it.\n1. Add the new script name to the `MIGRATION_SCRIPTS` constant at the top of\n   `edfi_lms_ds_loader/migrator.py`.\n\n### Adding New Files Uploads\n\n1. Create the required table and staging table in a new migration.\n1. Ensure that the `file-utils` shared library correctly maps the data types for\n   the new file.\n1. Update the `edfi_lms_ds_loader/loader_facade.py` to pull in the additional\n   file type and upload it.\n\n## Legal Information\n\nCopyright (c) 2021 Ed-Fi Alliance, LLC and contributors.\n\nLicensed under the [Apache License, Version 2.0](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/LICENSE) (the "License").\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.\n\nSee [NOTICES](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/NOTICES.md) for\nadditional copyright and license notifications.\n',
    'author': 'Ed-Fi Alliance, LLC and contributors',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://techdocs.ed-fi.org/display/EDFITOOLS/LMS+Toolkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List, Set

import pandas as pd
from sqlalchemy.engine.result import ResultProxy as sa_Result
from sqlalchemy.engine import Engine as sa_Engine
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import Session as sa_Session

from edfi_lms_ds_loader.helpers.constants import Table
from edfi_lms_ds_loader.sql_adapter import execute_transaction

logger = logging.getLogger(__name__)


class MssqlLmsOperations:
    """
    An adapter providing Microsoft SQL Server operations for management and use
    of LMS staging and production tables.

    Parameters
    ----------
    engine: sqlalchemy.engine.Engine
        SQL Alchemy engine.
    """

    engine: sa_Engine

    def __init__(self, engine: sa_Engine) -> None:
        self.engine = engine

    def _exec(self, statement: str) -> int:
        """This is a wrapper function that will not be unit tested."""

        assert statement.strip() != "", "Argument `statement` cannot be whitespace"

        def __callback(session: sa_Session) -> sa_Result:
            result: sa_Result = session.execute(statement)
            return result

        result = execute_transaction(self.engine, __callback)

        if result:
            return int(result.rowcount)

        return 0

    def truncate_staging_table(self, table: str) -> None:
        """
        Executes a truncate command on the staging version of a table.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        # Note: for postgresql we'll want `TRUNCATE TABLE {staging} RESTART IDENTITY`
        self._exec(f"TRUNCATE TABLE lms.stg_{table};")

    def disable_staging_natural_key_index(self, table: str) -> None:
        """
        Disables the natural key index on the staging table, for optimizing
        inserts.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        self._exec(
            f"ALTER INDEX IX_stg_{table}_Natural_Key on lms.stg_{table} DISABLE;"
        )

    def enable_staging_natural_key_index(self, table: str) -> None:
        """
        Re-builds the natural key index on the staging table.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        self._exec(
            f"ALTER INDEX IX_stg_{table}_Natural_Key on lms.stg_{table} REBUILD;"
        )

    def insert_into_staging(self, df: pd.DataFrame, table: str) -> None:
        """
        Inserts all records from a DataFrame into the staging table.

        Parameters
        ----------
        df: DataFrame
            A Pandas dataframe with column names that match the destination table
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        df.to_sql(
            f"stg_{table}",
            self.engine,
            schema="lms",
            if_exists="append",
            index=False,
            method="multi",
            chunksize=120,
        )
        logger.debug(f"All records have been loaded into staging table 'stg_{table}'")

    def insert_new_records_to_production(self, table: str, columns: List[str]) -> None:
        """
        Copies new records from the staging table to the production table.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        columns: List[str]
            A list of the column names in the table
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        column_string = ",".join([f"\n    {c}" for c in columns])

        statement = f"""
INSERT INTO
    lms.{table}
({column_string}
)
SELECT{column_string}
FROM
    lms.stg_{table} as stg
WHERE
    NOT EXISTS (
        SELECT
            1
        FROM
            lms.{table}
        WHERE
            SourceSystemIdentifier = stg.SourceSystemIdentifier
        AND
            SourceSystem = stg.SourceSystem
    )
"""

        row_count = self._exec(statement)
        logger.debug(f"Inserted {row_count} records into table `{table}`.")

    def insert_new_records_to_production_for_user_relation(
        self, table: str, columns: List[str]
    ) -> None:
        """
        Copies new records from the staging table to the production table. Specialized
        for tables that have a foreign key to LMSUser.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        columns: List[str]
            A list of the column names in the table
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        insert_columns = ",".join(
            [f"\n    {c}" for c in columns if c != "LMSUserSourceSystemIdentifier"]
        )
        select_columns = ",".join(
            [f"\n    stg.{c}" for c in columns if c != "LMSUserSourceSystemIdentifier"]
        )

        statement = f"""
INSERT INTO
    lms.{table}
(
    LMSUserIdentifier,{insert_columns}
)
SELECT
    LMSUser.LMSUserIdentifier,{select_columns}
FROM
    lms.stg_{table} as stg
INNER JOIN
    lms.LMSUser
ON
    stg.LMSUserSourceSystemIdentifier = LMSUser.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSUser.SourceSystem
WHERE NOT EXISTS (
  SELECT
    1
  FROM
    lms.{table}
  WHERE
    SourceSystemIdentifier = stg.SourceSystemIdentifier
  AND
    SourceSystem = stg.SourceSystem
)
"""

        row_count = self._exec(statement)
        logger.debug(f"Inserted {row_count} records into table `{table}`.")

    def insert_new_records_to_production_for_section_relation(
        self, table: str, columns: List[str]
    ) -> None:
        """
        Copies new records from the staging table to the production table. Specialized
        for tables that have a foreign key to LMSSection.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        columns: List[str]
            A list of the column names in the table
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        insert_columns = ",".join(
            [f"\n    {c}" for c in columns if c != "LMSSectionSourceSystemIdentifier"]
        )
        select_columns = ",".join(
            [
                f"\n    stg.{c}"
                for c in columns
                if c != "LMSSectionSourceSystemIdentifier"
            ]
        )

        statement = f"""
INSERT INTO
    lms.{table}
(
    LMSSectionIdentifier,{insert_columns}
)
SELECT
    LMSSection.LMSSectionIdentifier,{select_columns}
FROM
    lms.stg_{table} as stg
INNER JOIN
    lms.LMSSection
ON
    stg.LMSSectionSourceSystemIdentifier = LMSSection.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSSection.SourceSystem
WHERE NOT EXISTS (
  SELECT
    1
  FROM
    lms.{table}
  WHERE
    SourceSystemIdentifier = stg.SourceSystemIdentifier
  AND
    SourceSystem = stg.SourceSystem
)
"""

        row_count = self._exec(statement)
        logger.debug(f"Inserted {row_count} records into table `{table}`.")

    def insert_new_records_to_production_for_section_and_user_relation(
        self, table: str, columns: List[str]
    ) -> None:
        """
        Copies new records from the staging table to the production table. Specialized
        for tables that have a foreign key to both LMSSection and LMSUser.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        columns: List[str]
            A list of the column names in the table
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        insert_columns = ",".join(
            [
                f"\n    {c}"
                for c in columns
                if not (
                    c == "LMSSectionSourceSystemIdentifier"
                    or c == "LMSUserSourceSystemIdentifier"
                )
            ]
        )
        select_columns = ",".join(
            [
                f"\n    stg.{c}"
                for c in columns
                if not (
                    c == "LMSSectionSourceSystemIdentifier"
                    or c == "LMSUserSourceSystemIdentifier"
                )
            ]
        )

        statement = f"""
INSERT INTO
    lms.{table}
(
    LMSSectionIdentifier,
    LMSUserIdentifier,{insert_columns}
)
SELECT
    LMSSection.LMSSectionIdentifier,
    LMSUser.LMSUserIdentifier,{select_columns}
FROM
    lms.stg_{table} as stg
INNER JOIN
    lms.LMSSection
ON
    stg.LMSSectionSourceSystemIdentifier = LMSSection.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSSection.SourceSystem
INNER JOIN
    lms.LMSUser
ON
    stg.LMSUserSourceSystemIdentifier = LMSUser.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSUser.SourceSystem
WHERE NOT EXISTS (
  SELECT
    1
  FROM
    lms.{table}
  WHERE
    SourceSystemIdentifier = stg.SourceSystemIdentifier
  AND
    SourceSystem = stg.SourceSystem
)
"""

        row_count = self._exec(statement)
        logger.debug(f"Inserted {row_count} records into table `{table}`.")

    def insert_new_records_to_production_for_assignment_and_user_relation(
        self, table: str, columns: List[str]
    ) -> None:
        """
        Copies new records from the staging table to the production table. Specialized
        for tables that have a foreign key to both Assignment and LMSUser.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        columns: List[str]
            A list of the column names in the table
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        insert_columns = ",".join(
            [
                f"\n    {c}"
                for c in columns
                if not (
                    c == "AssignmentSourceSystemIdentifier"
                    or c == "LMSUserSourceSystemIdentifier"
                )
            ]
        )
        select_columns = ",".join(
            [
                f"\n    stg.{c}"
                for c in columns
                if not (
                    c == "AssignmentSourceSystemIdentifier"
                    or c == "LMSUserSourceSystemIdentifier"
                )
            ]
        )

        statement = f"""
INSERT INTO
    lms.{table}
(
    AssignmentIdentifier,
    LMSUserIdentifier,{insert_columns}
)
SELECT
    Assignment.AssignmentIdentifier,
    LMSUser.LMSUserIdentifier,{select_columns}
FROM
    lms.stg_{table} as stg
INNER JOIN
    lms.Assignment
ON
    stg.AssignmentSourceSystemIdentifier = Assignment.SourceSystemIdentifier
AND
    stg.SourceSystem = Assignment.SourceSystem
INNER JOIN
    lms.LMSUser
ON
    stg.LMSUserSourceSystemIdentifier = LMSUser.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSUser.SourceSystem
WHERE NOT EXISTS (
  SELECT
    1
  FROM
    lms.{table}
  WHERE
    SourceSystemIdentifier = stg.SourceSystemIdentifier
  AND
    SourceSystem = stg.SourceSystem
)
"""

        row_count = self._exec(statement)
        logger.debug(f"Inserted {row_count} records into table `{table}`.")

    def insert_new_records_to_production_for_attendance_events(
        self, table: str, columns: List[str]
    ) -> None:
        """
        Copies new records from the staging table to the production table. Specialized
        for the Attendance Events table.

        Parameters
        ----------
        table: str
            Not strictly necessary, but must be part of the signature
        columns: List[str]
            A list of the column names in the table
        """
        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        def __not_a_foreign_key(column: str) -> bool:
            return column not in [
                "LMSSectionSourceSystemIdentifier",
                "LMSUserSourceSystemIdentifier",
            ]

        insert_columns = ",".join(
            [f"\n    {c}" for c in columns if __not_a_foreign_key(c)]
        )
        select_columns = ",".join(
            [f"\n    stg.{c}" for c in columns if __not_a_foreign_key(c)]
        )

        statement = f"""
INSERT INTO
    lms.LMSUserAttendanceEvent
(
    LMSSectionIdentifier,
    LMSUserIdentifier,
    LMSUserLMSSectionAssociationIdentifier,{insert_columns}
)
SELECT
    LMSSection.LMSSectionIdentifier,
    LMSUser.LMSUserIdentifier,
    LMSUserLMSSectionAssociation.LMSUserLMSSectionAssociationIdentifier,{select_columns}
FROM
    lms.stg_LMSUserAttendanceEvent as stg
INNER JOIN
    lms.LMSSection
ON
    stg.LMSSectionSourceSystemIdentifier = LMSSection.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSSection.SourceSystem
INNER JOIN
    lms.LMSUser
ON
    stg.LMSUserSourceSystemIdentifier = LMSUser.SourceSystemIdentifier
AND
    stg.SourceSystem = LMSUser.SourceSystem
INNER JOIN
    lms.LMSUserLMSSectionAssociation
ON
    LMSUser.LMSUserIdentifier = LMSUserLMSSectionAssociation.LMSUserIdentifier
AND
    LMSSection.LMSSectionIdentifier = LMSUserLMSSectionAssociation.LMSSectionIdentifier
WHERE NOT EXISTS (
  SELECT
    1
  FROM
    lms.LMSUserAttendanceEvent
  WHERE
    SourceSystemIdentifier = stg.SourceSystemIdentifier
  AND
    SourceSystem = stg.SourceSystem
)
"""

        row_count = self._exec(statement)
        logger.debug(f"Inserted {row_count} records into table `{table}`.")

    def copy_updates_to_production(self, table: str, columns: List[str]) -> None:
        """
        Updates modified records in production based on the staging table, based
        on the LastModifiedDate.

        Parameters
        ----------
        table: str
            Name of the table to truncate, not including the `stg_` prefix
        columns: List[str]
            A list of the column names in the table
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"
        assert len(columns) > 0, "Argument `columns` cannot be empty"

        update_columns = (
            ",".join(
                [
                    f"\n    {c} = stg.{c}"
                    for c in columns
                    if c
                    not in (
                        # These are natural key columns that should never be
                        # updated.
                        "SourceSystem",
                        "SourceSystemIdentifier",
                        "LMSSectionSourceSystemIdentifier",
                        "LMSUserSourceSystemIdentifier",
                        "AssignmentSourceSystemIdentifier",
                    )
                ]
            )
            + ",\n    DeletedAt = NULL"
        )

        statement = f"""
UPDATE
    t
SET{update_columns}
FROM
    lms.{table} as t
INNER JOIN
    lms.stg_{table} as stg
ON
    t.SourceSystem = stg.SourceSystem
AND
    t.SourceSystemIdentifier = stg.SourceSystemIdentifier
AND
    t.LastModifiedDate <> stg.LastModifiedDate
"""

        row_count = self._exec(statement)
        logger.debug(f"Updated {row_count} records in table `{table}`.")

    def soft_delete_from_production(self, table: str, source_system: str) -> None:
        """
        Updates production records that do not have a match in the staging table
        by setting their `deletedat` value to the current timestamp.

        Parameters
        ----------
        table: str
            Name of the table to soft delete on, not including the `stg_`.
        source_system: str
            The SourceSystem currently being processed.
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        statement = f"""
UPDATE
    t
SET
    t.DeletedAt = getdate()
FROM
    lms.{table} as t
WHERE
    NOT EXISTS (
        SELECT
            1
        FROM
            lms.stg_{table} as stg
        WHERE
            t.SourceSystemIdentifier = stg.SourceSystemIdentifier
        AND
            t.SourceSystem = stg.SourceSystem
    )
AND
    t.DeletedAt IS NULL
AND
    t.SourceSystem = '{source_system}'
"""

        row_count = self._exec(statement)
        logger.debug(f"Soft-deleted {row_count} records in table `{table}`")

    def soft_delete_from_production_for_section_relation(
        self, table: str, source_system: str
    ) -> None:
        """
        Updates production records that do not have a match in the staging table
        by setting their `deletedat` value to the current timestamp, but
        only for the related sections in the staging table.

        Parameters
        ----------
        table: str
            Name of the table to soft delete on, not including the `stg_`.
        source_system: str
            The SourceSystem currently being processed.
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        statement = f"""
UPDATE
    t
SET
    t.DeletedAt = getdate()
FROM
    lms.{table} as t
WHERE
    t.LMSSectionIdentifier IN (
        SELECT
            s.LMSSectionIdentifier
        FROM
           lms.LMSSection as s
        INNER JOIN
            lms.stg_{table} as stg
        ON
            stg.LMSSectionSourceSystemIdentifier = s.SourceSystemIdentifier
        AND
            stg.SourceSystem = s.SourceSystem
    )
AND
    NOT EXISTS (
        SELECT
            1
        FROM
            lms.stg_{table} as stg
        WHERE
            t.SourceSystemIdentifier = stg.SourceSystemIdentifier
        AND
            t.SourceSystem = stg.SourceSystem
    )
AND
    t.DeletedAt IS NULL
AND
    t.SourceSystem = '{source_system}'
"""

        row_count = self._exec(statement)
        logger.debug(f"Soft-deleted {row_count} records in table `{table}`")

    def soft_delete_from_production_for_assignment_relation(
        self, table: str, source_system: str
    ) -> None:
        """
        Updates production records that do not have a match in the staging table
        by setting their `deletedat` value to the current timestamp, but
        only for the related assignments in the staging table.

        Parameters
        ----------
        table: str
            Name of the table to soft delete on, not including the `stg_`.
        source_system: str
            The SourceSystem currently being processed.
        """

        assert table.strip() != "", "Argument `table` cannot be whitespace"

        statement = f"""
UPDATE
    t
SET
    t.DeletedAt = getdate()
FROM
    lms.{table} as t
WHERE
    t.AssignmentIdentifier IN (
        SELECT
            a.AssignmentIdentifier
        FROM
           lms.Assignment as a
        INNER JOIN
            lms.stg_{table} as stg
        ON
            stg.AssignmentSourceSystemIdentifier = a.SourceSystemIdentifier
        AND
            stg.SourceSystem = a.SourceSystem
    )
AND
    NOT EXISTS (
        SELECT
            1
        FROM
            lms.stg_{table} as stg
        WHERE
            t.SourceSystemIdentifier = stg.SourceSystemIdentifier
        AND
            t.SourceSystem = stg.SourceSystem
    )
AND
    t.DeletedAt IS NULL
AND
    t.SourceSystem = '{source_system}'
"""

        row_count = self._exec(statement)
        logger.debug(f"Soft-deleted {row_count} records in table `{table}`")

    def insert_new_submission_types(self) -> None:
        """
        Inserts new Assignment Submission Type records from staging table
        into the production table.
        """

        statement = """
INSERT INTO lms.AssignmentSubmissionType (
    AssignmentIdentifier,
    SubmissionType
)
SELECT
    Assignment.AssignmentIdentifier,
    stg_AssignmentSubmissionType.SubmissionType
FROM
        lms.stg_AssignmentSubmissionType
    INNER JOIN
        lms.Assignment
    ON
        stg_AssignmentSubmissionType.SourceSystem = Assignment.SourceSystem
    AND
        stg_AssignmentSubmissionType.SourceSystemIdentifier = Assignment.SourceSystemIdentifier
WHERE
    NOT EXISTS (
        SELECT
            1
        FROM
            lms.AssignmentSubmissionType
        WHERE
            AssignmentIdentifier = Assignment.AssignmentIdentifier
        AND
            SubmissionType = stg_AssignmentSubmissionType.SubmissionType
    )
"""

        row_count = self._exec(statement)
        logger.debug(
            f"Updated {row_count} records in table `{Table.ASSIGNMENT_SUBMISSION_TYPES}`."
        )

    def soft_delete_removed_submission_types(self, source_system: str) -> None:
        """
        Marks existing Assignment Submission Types as "deleted" when they are
        no longer present in the incoming data.

        Parameters
        ----------
        source_system: str
            The name of the source system for the current import process.
        """

        statement = f"""
UPDATE
    AssignmentSubmissionType
SET
    DeletedAt = GETDATE()
FROM
    lms.AssignmentSubmissionType
INNER JOIN
    lms.Assignment
ON
    AssignmentSubmissionType.AssignmentIdentifier = Assignment.AssignmentIdentifier
WHERE
    SourceSystem = '{source_system}'
AND
    NOT EXISTS (
        SELECT
            1
        FROM
            lms.stg_AssignmentSubmissionType
        WHERE
            stg_AssignmentSubmissionType.SourceSystem = Assignment.SourceSystem
        AND
            stg_AssignmentSubmissionType.SourceSystemIdentifier = Assignment.SourceSystemIdentifier
        AND
            stg_AssignmentSubmissionType.SubmissionType = AssignmentSubmissionType.SubmissionType
    )
"""
        row_count = self._exec(statement)
        logger.debug(
            f"Soft deleted {row_count} records in table `{Table.ASSIGNMENT_SUBMISSION_TYPES}`."
        )

    def unsoft_delete_returned_submission_types(self, source_system: str) -> None:
        """
        Unmarks previously "deleted" Assignment Submission Types when they are
        present in the incoming data.

        Parameters
        ----------
        source_system: str
            The name of the source system for the current import process.
        """

        statement = f"""
UPDATE
    AssignmentSubmissionType
SET
    DeletedAt = NULL
FROM
    lms.AssignmentSubmissionType
INNER JOIN
    lms.Assignment
ON
    AssignmentSubmissionType.AssignmentIdentifier = Assignment.AssignmentIdentifier
WHERE
    SourceSystem = '{source_system}'
AND
    EXISTS (
        SELECT
            1
        FROM
            lms.stg_AssignmentSubmissionType
        WHERE
            stg_AssignmentSubmissionType.SourceSystem = Assignment.SourceSystem
        AND
            stg_AssignmentSubmissionType.SourceSystemIdentifier = Assignment.SourceSystemIdentifier
        AND
            stg_AssignmentSubmissionType.SubmissionType = AssignmentSubmissionType.SubmissionType
    )
"""
        row_count = self._exec(statement)
        logger.debug(
            f"Un-soft deleted {row_count} records in table `{Table.ASSIGNMENT_SUBMISSION_TYPES}`."
        )

    def get_processed_files(self, resource_name: str) -> Set[str]:
        try:
            query = f"""
SELECT
    FullPath
FROM
    lms.ProcessedFiles
WHERE
    ResourceName = '{resource_name}'
""".strip()
            result = pd.read_sql_query(query, self.engine)
            if "FullPath" in result:
                return set(result["FullPath"])
            return set()
        except ProgrammingError as pe:
            logger.exception(pe)
            raise

    def add_processed_file(self, path: str, resource_name: str, rows: int):
        """
        Records that a file has been processed and thus should not be processed
        a second time.

        Parameters
        ----------
        path: str
            Filesystem path for the file that was processed.
        resource_name: str
            Name of the resource covered by the file.
        rows: int
            Number of rows in the file.
        """

        statement = f"""
INSERT INTO
    lms.ProcessedFiles
(
    FullPath,
    ResourceName,
    NumberOfRows
)
VALUES
(
    '{path}',
    '{resource_name}',
    {rows}
)
""".strip()

        try:
            _ = self._exec(statement)
        except ProgrammingError as pe:
            logger.exception(pe)
            raise
